/*
    UPN-Calc - An arbitrary precision RPN calculator for J2ME/MIDP1
    Copyright (C) 2005 Kai G. Schwebke, upn.schwebke.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 * Arbitrary Precision BCD Integer Arithmetic
 */

class BCDInteger
{

   byte[] d;         // number: sum(i = 0 ... numDigits-1; d[i]*10^i)
   boolean neg;      // sign
   int numDigits;    // number of digits
   BCDInteger lastR; // remainder of last long division
   byte lastRshort;  // remainder of last short division

   // new Number, val = 0, length = 0
   public BCDInteger()
   {
      initDigits(0);
   }

   // new Number, val = 0, length = n
   public BCDInteger(int n)
   {
      initDigits(n);
   }

   // new Number, val and length from string
   public BCDInteger(String s)
   {
      initFromString(s);
   }

   public BCDInteger copy()
   {
      BCDInteger r = new BCDInteger(numDigits);
      r.neg = neg;

      int i;
      for (i = 0; i < numDigits; ++i)
      {
	 r.d[i] = d[i];
      }

      return r;
   }

   // signed compare
   public int cmp(BCDInteger a)
   {
      if ( (!neg) && (!a.neg) )
      {
	 return cmpUnsigned(a);
      }
      if (neg && a.neg)
      {
	 return -cmpUnsigned(a);
      }
      if (neg && (!a.neg))
      {
	 return -1;
      }
      return 1;
   }

   // unsigned (absolute) compare
   public int cmpUnsigned(BCDInteger a)
   {
      int cmpDigits = Math.max(numDigits, a.numDigits);

      int i;
      for (i = cmpDigits-1; i >= 0; --i)
      {
	 int digit, digita;
	 digit = digita = 0;

	 if (i < numDigits)
	 {
	    digit = d[i];
	 }
	 if (i < a.numDigits)
	 {
	    digita = a.d[i];
	 }
	 if (digit > digita)
	 {
	    return 1;
	 }
	 if (digit < digita)
	 {
	    return -1;
	 }
      }
      return 0;
   }

   // signed addition
   public BCDInteger add(BCDInteger a)
   {
      if (neg == a.neg)
      {
	 BCDInteger r = addUnsigned(a);
	 return r;
      }

      BCDInteger r;
      int c = cmpUnsigned(a);
      if (c > 0)
      {
	 r = subUnsigned(a);
      } else {
	 r = a.subUnsigned(this);
      }

      if ( ( (c > 0) && (neg) ) || ( (c < 0) && (a.neg) ) )
      {
	 r.neg = true;
      }

      return r;
   }

   // signed subtraction
   public BCDInteger sub(BCDInteger at)
   {
      BCDInteger a = at.copy();
      a.neg = !a.neg;
      BCDInteger r = add(a);
      a.neg = !a.neg;
      return r;
   }

   // unsigned (absolute) addition
   public BCDInteger addUnsigned(BCDInteger a)
   {
      int newDigits = Math.max(numDigits, a.numDigits)+1;
      BCDInteger newInt = new BCDInteger(newDigits);
      
      int i;
      for (i = 0; i < (newDigits-1); ++i)
      {
	 if (i < numDigits)
	 {
	    newInt.d[i] += d[i];
	 }
	 if (i < a.numDigits)
	 {
	    newInt.d[i] += a.d[i];
	 }
	 if (newInt.d[i] >= 10)
	 {
	    newInt.d[i] -= 10;
	    ++newInt.d[i+1];
	 }
      }

      newInt.neg = neg;

      return newInt;
   }

   // unsigned (absolute) addition of one digit
   public BCDInteger addUnsigned(int a)
   {
      int newDigits = numDigits+1;
      BCDInteger newInt = new BCDInteger(newDigits);
      
      newInt.d[0] = (byte)a;

      int i;
      for (i = 0; i < (newDigits-1); ++i)
      {
	 if (i < numDigits)
	 {
	    newInt.d[i] += d[i];
	 }
	 if (newInt.d[i] >= 10)
	 {
	    newInt.d[i] -= 10;
	    ++newInt.d[i+1];
	 }
      }

      newInt.neg = neg;

      return newInt;
   }

   // unsigned (absolute) subtraction
   // predicate: this > a
   //            (or this < a if numDigits == a.numDigits and
   //             underflow in most significant digit is allowed)
   public BCDInteger subUnsigned(BCDInteger a)
   {
      int newDigits = numDigits;
      BCDInteger newInt = new BCDInteger(newDigits);
      
      int i;
      for (i = 0; i < numDigits; ++i)
      {
	 newInt.d[i] += d[i];
	 if (i < a.numDigits)
	 {
	    newInt.d[i] -= a.d[i];
	 }
	 if ( (newInt.d[i] < 0) && (i < (numDigits-1)) )
	 {
	    newInt.d[i] += 10;
	    --newInt.d[i+1];
	 }
      }

      return newInt;
   }

   // signed multiplication
   public BCDInteger mul(BCDInteger a)
   {
      BCDInteger p = mulUnsigned(a);

      if (neg != a.neg)
      {
	 p.neg = true;
      }

      return p;
   }

   // unsigned (absolute) multiplication
   public BCDInteger mulUnsigned(BCDInteger a)
   {
      int newDigits = numDigits+a.numDigits;
      BCDInteger newInt = new BCDInteger(0);

      int i;
      for (i = 0; i < a.numDigits; ++i)
      {
	 newInt = newInt.addUnsigned(this.shiftRight(i).mulUnsigned(a.d[i]));
      }

      return newInt;
   }

   // unsigned (absolute) multiplication by one digit
   public BCDInteger mulUnsigned(int digit)
   {
      int newDigits = numDigits+1;
      BCDInteger newInt = new BCDInteger(newDigits);

      int i;
      for (i = 0; i < numDigits; ++i)
      {
	 int t = newInt.d[i] + digit*d[i];
	 newInt.d[i] = (byte)(t%10);
	 newInt.d[i+1] = (byte)(t/10);
      }

      return newInt;
   }

   // signed division
   public BCDInteger div(BCDInteger v)
   {
      BCDInteger q;

      v.shorten();
      if (v.numDigits > 1)
      {
	 q = divUnsigned(v);
      } else {
	 q = divUnsigned(v.d[0]);
	 lastR = new BCDInteger(1);
	 lastR.d[0] = lastRshort;
      }

      if (neg != v.neg)
      {
	 q.neg = true;
      }

      return q;
   }

   // predicate: numDigits > v.numDigits,
   //            v = v.shorten(),
   //            v.numDigits > 1
   public BCDInteger divUnsigned(BCDInteger v)
   {
      int n = v.numDigits;
      int m = numDigits-n;

      BCDInteger q = new BCDInteger(m+1);

      byte d = (byte)(10/(v.d[n-1]+1));
      BCDInteger u = mulUnsigned(d);
      v = v.mulUnsigned(d);

      int j;
      byte qc, rc;
      for (j = m; j >= 0; --j)
      {
	 qc = (byte)((u.d[j+n]*10+u.d[j+n-1])/v.d[n-1]);
	 rc = (byte)((u.d[j+n]*10+u.d[j+n-1])%v.d[n-1]);
	 if ( (qc == 10) || (qc*v.d[n-2] > 10*rc+u.d[j+n-2]) )
	 {
	    --qc;
	    rc += v.d[n-1];

	    if (rc < 10)
	    {
	       if ( (qc == 10) || (qc*v.d[n-2] > 10*rc+u.d[j+n-2]) )
	       {
		  --qc;
		  rc += v.d[n-1];
	       }
	    }
	 }

	 BCDInteger ut = u.mid(j, n+1);
	 ut = ut.sub(v.mulUnsigned(qc));

	 q.d[j] = qc;
	 if (ut.neg)
	 {
	    --q.d[j];
	    ut = ut.add(v);
	 }

	 int k;
	 for (k = 0; k <= n; ++k)
	 {
	    u.d[k+j] = ut.d[k];
	 }
      }

      lastR = u.divUnsigned(d);

      return q;
   }

   public BCDInteger divUnsigned(int v)
   {
      BCDInteger u = this.mulUnsigned(1);
      BCDInteger q = new BCDInteger(numDigits);

      int i;
      for (i = numDigits-1; i >= 0; --i)
      {
	 int ut = u.d[i+1]*10+u.d[i];
	 q.d[i] = (byte)(ut/v);
	 u.d[i+1] -= (byte)(q.d[i]*v/10);
	 u.d[i] -= (byte)(q.d[i]*v%10);
	 if (u.d[i] < 0)
	 {
	    u.d[i] += 10;
	    --u.d[i+1];
	 }
      }

      lastRshort = u.d[0];

      return q;
   }

   // shift right by n digits (= multiply by 10^n)
   public BCDInteger shiftRight(int n)
   {
      int newDigits = numDigits+n;
      BCDInteger newInt = new BCDInteger(newDigits);
      newInt.neg = neg;

      int i;
      for (i = n; i < newDigits; ++i)
      {
	 newInt.d[i] = d[i-n];
      }

      return newInt;
   }

   // shift left by n digits (= divide by 10^n)
   public BCDInteger shiftLeft(int n)
   {
      int newDigits = numDigits-n;
      BCDInteger newInt = new BCDInteger(newDigits);
      newInt.neg = neg;

      int i;
      for (i = 0; i < newDigits; ++i)
      {
	 newInt.d[i] = d[i+n];
      }

      return newInt;
   }


   // shorten (remove leading zero digits)
   public void shorten()
   {
      int sh = numDigits-1;
      while ((sh >= 0) && (d[sh] == 0))
      {
	 --sh;
      }

      if (sh < (numDigits-1))
      {
	 if (sh == (-1))
	 {
	    initDigits(0);
	 } else {
	    byte[] nd = new byte[sh+1];

	    int i;
	    for (i = 0; i <= sh; ++i)
	    {
	       nd[i] = d[i];
	    }

	    d = nd;
	    numDigits = sh+1;
	 }
      }
   }

   // get part of number
   public BCDInteger mid(int start, int len)
   {
      BCDInteger newInt = new BCDInteger(len);

      int i;
      for (i = start; i < start+len; ++i)
      {
	 newInt.d[i-start] = d[i];
      }

      return newInt;
   }

   // set val = 0 and numDigits = n
   public void initDigits(int n)
   {
      numDigits = n;
      if (numDigits > 0)
      {
	 d = new byte[numDigits];
      } else {
	 numDigits = 0;
	 d = null;
      }
      neg = false;
   }

   // return numDigits
   public int digits()
   {
      return numDigits;
   }

   // set val and length from string
   public void initFromString(String s)
   {
      numDigits = s.length();

      int i = 0;
      neg = false;

      if (s.charAt(0) == '-')
      {
	 neg = true;
      }

      while ((i < s.length()) && 
	     ((s.charAt(i) == '+') || (s.charAt(i) == '-') || (s.charAt(i) == '0')))
      {
	 ++i;
	 --numDigits;
      }

      if (numDigits <= 0)
      {
	 numDigits = 0;
	 d = null;
	 return;
      }

      int j = numDigits-1;
      d = new byte[numDigits];
      while (i < s.length())
      {
	 d[j] = (byte)Character.digit(s.charAt(i), 10);
	 --j;
	 ++i;
      }
   }

   // set val and length from integer
   public void initFromInteger(int i)
   {
      initFromString(Integer.toString(i));
   }

   // return string representation of number
   public String toString()
   {
      String s = "";
      if (neg)
      {
	 s = "-";
      }

      if (numDigits == 0)
      {
	 s = s + "0";
      } else {
	 boolean leadingZeros = true;
	 int i;
	 for (i = numDigits-1; i >= 0; --i)
	 {
	    if ((d[i] != 0) || (i == 0))
	    {
	       leadingZeros = false;
	    }
	    if (!leadingZeros)
	    {
	       s = s + Integer.toString((int)d[i], 10);
	    }
	 }
      }

      return s;
   }

   public int toInteger()
   {
      try {
	 return Integer.parseInt(toString());
      } catch (NumberFormatException e) {
      }
      return 0;
   }

}
